/*
 Copyright (c) 2013, Joel Levin
 All rights reserved.

 Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

 Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
 Neither the name of JLRoutes nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "JLRoutes.h"


NSString *const kJLRoutePatternKey = @"JLRoutePattern";
NSString *const kJLRouteURLKey = @"JLRouteURL";
NSString *const kJLRouteNamespaceKey = @"JLRouteNamespace";
NSString *const kJLRouteWildcardComponentsKey = @"JLRouteWildcardComponents";

NSString *const kJLRoutesGlobalNamespaceKey = @"JLRoutesGlobalNamespace";


static NSMutableDictionary *routeControllersMap = nil;
static BOOL verboseLoggingEnabled = NO;
static BOOL shouldDecodePlusSymbols = YES;


@interface JLRoutes ()

@property (nonatomic, strong) NSMutableArray *routes;
@property (nonatomic, strong) NSString *namespaceKey;

@end


#pragma mark -

@interface _JLRoute : NSObject

@property (nonatomic, weak) JLRoutes *parentRoutesController;
@property (nonatomic, strong) NSString *pattern;
@property (nonatomic, strong) BOOL (^block)(NSDictionary *parameters);
@property (nonatomic, assign) NSUInteger priority;
@property (nonatomic, strong) NSArray *patternPathComponents;
@property (nonatomic, strong) NSArray *patternFragmentComponents;
@property (nonatomic, assign) BOOL matchFragmentComponents;

- (NSDictionary *)parametersForURL:(NSURL *)URL pathComponents:(NSArray *)pathComponents fragmentComponents:(NSArray *)fragmentComponents;

@end


#pragma mark -

@interface NSString (JLRoutes)

- (NSString *)JLRoutes_URLDecodedString;
- (NSDictionary *)JLRoutes_URLParameterDictionary;

@end

@interface NSURL (JLRoutes)

- (NSArray *)JLRoutes_fragmentPathComponents;
- (NSString *)JLRoutes_fragmentQuery;

@end


#pragma mark -

@implementation JLRoutes

- (instancetype)init
{
    if ((self = [super init])) {
        self.routes = [NSMutableArray array];
    }
    return self;
}

- (NSString *)description
{
    return [self.routes description];
}

+ (NSString *)allRoutes
{
    NSMutableString *descriptionString = [NSMutableString stringWithString:@"\n"];
    
    for (NSString *routesNamespace in routeControllersMap) {
        JLRoutes *routesController = routeControllersMap[routesNamespace];
        [descriptionString appendFormat:@"\"%@\":\n%@\n\n", routesController.namespaceKey, routesController.routes];
    }
    
    return descriptionString;
}


#pragma mark - Routing Schemes

+ (instancetype)globalRoutes
{
    return [self routesForScheme:kJLRoutesGlobalNamespaceKey];
}

+ (instancetype)routesForScheme:(NSString *)scheme
{
    JLRoutes *routesController = nil;
    
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        routeControllersMap = [[NSMutableDictionary alloc] init];
    });
    
    if (!routeControllersMap[scheme]) {
        routesController = [[self alloc] init];
        routesController.namespaceKey = scheme;
        routeControllersMap[scheme] = routesController;
    }
    
    routesController = routeControllersMap[scheme];
    
    return routesController;
}

+ (void)unregisterRouteScheme:(NSString *)scheme
{
    [routeControllersMap removeObjectForKey:scheme];
}

+ (void)unregisterAllRouteSchemes
{
    [routeControllersMap removeAllObjects];
}


#pragma mark - Registering Routes

- (void)addRoute:(NSString *)routePattern handler:(BOOL (^)(NSDictionary<NSString *, id> *parameters))handlerBlock
{
    [self addRoute:routePattern priority:0 handler:handlerBlock];
}

- (void)addRoutes:(NSArray<NSString *> *)routePatterns handler:(BOOL (^)(NSDictionary<NSString *, id> *parameters))handlerBlock
{
    for (NSString *routePattern in routePatterns) {
        [self addRoute:routePattern handler:handlerBlock];
    }
}

- (void)addRoute:(NSString *)routePattern priority:(NSUInteger)priority handler:(BOOL (^)(NSDictionary<NSString *, id> *parameters))handlerBlock
{
    // if there's a pair of parenthesis, process optionals, trim the parenthesis, put it on trimmedRoute
    NSString *trimmedRoute = routePattern;
    
    // repeat until no parenthesis pair is found
    while ([trimmedRoute rangeOfString:@")" options:NSBackwardsSearch].location > [trimmedRoute rangeOfString:@"(" options:NSBackwardsSearch].location) {
        // build route with the optionals
        NSString *patternWithOptionals = [trimmedRoute stringByReplacingOccurrencesOfString:@"(" withString:@""];
        patternWithOptionals = [patternWithOptionals stringByReplacingOccurrencesOfString:@")" withString:@""];
        [self _registerRoute:patternWithOptionals priority:priority handler:handlerBlock];
        
        // build route without optionals
        NSRange rangeOfLastParentheses = [trimmedRoute rangeOfString:@"(" options:NSBackwardsSearch];
        NSRange rangeToRemove = NSMakeRange(rangeOfLastParentheses.location, trimmedRoute.length - rangeOfLastParentheses.location);
        NSString *patternWithLastOptionalRemoved = [trimmedRoute stringByReplacingCharactersInRange:rangeToRemove withString:@""];
        
        // remove any parenthesis for other optionals that might still be in the route
        NSString *patternWithoutOptionals = [patternWithLastOptionalRemoved stringByReplacingOccurrencesOfString:@"(" withString:@""];
        patternWithoutOptionals = [patternWithoutOptionals stringByReplacingOccurrencesOfString:@")" withString:@""];
        [self _registerRoute:patternWithoutOptionals priority:priority handler:handlerBlock];
        
        trimmedRoute = patternWithLastOptionalRemoved;
    }
    
    //Only register original route if trimmedRoute haven't been modified.
    if (trimmedRoute == routePattern) {
        [self _registerRoute:routePattern priority:priority handler:handlerBlock];
    }
}

- (void)removeRoute:(NSString *)routePattern
{
    if (![routePattern hasPrefix:@"/"]) {
        routePattern = [NSString stringWithFormat:@"/%@", routePattern];
    }
    
    NSInteger routeIndex = NSNotFound;
    NSInteger index = 0;
    
    for (_JLRoute *route in self.routes) {
        if ([route.pattern isEqualToString:routePattern]) {
            routeIndex = index;
            break;
        }
        index++;
    }
    
    if (routeIndex != NSNotFound) {
        [self.routes removeObjectAtIndex:(NSUInteger)routeIndex];
    }
}

- (void)removeAllRoutes
{
    [self.routes removeAllObjects];
}

- (void)setObject:(id)handlerBlock forKeyedSubscript:(NSString *)routePatten
{
    [self addRoute:routePatten handler:handlerBlock];
}


#pragma mark - Routing URLs

+ (BOOL)canRouteURL:(NSURL *)URL
{
    return [[self _routesControllerForURL:URL] canRouteURL:URL];
}

- (BOOL)canRouteURL:(NSURL *)URL
{
    return [self _routeURL:URL withParameters:nil executeRouteBlock:NO];
}

+ (BOOL)routeURL:(NSURL *)URL
{
    return [[self _routesControllerForURL:URL] routeURL:URL];
}

- (BOOL)routeURL:(NSURL *)URL
{
    return [self _routeURL:URL withParameters:nil executeRouteBlock:YES];
}

+ (BOOL)routeURL:(NSURL *)URL withParameters:(NSDictionary *)parameters
{
    return [[self _routesControllerForURL:URL] routeURL:URL withParameters:parameters];
}

- (BOOL)routeURL:(NSURL *)URL withParameters:(NSDictionary *)parameters
{
    return [self _routeURL:URL withParameters:parameters executeRouteBlock:YES];
}


#pragma mark - Private

- (void)_registerRoute:(NSString *)routePattern priority:(NSUInteger)priority handler:(BOOL (^)(NSDictionary *parameters))handlerBlock
{
    _JLRoute *route = [[_JLRoute alloc] init];
    route.pattern = routePattern;
    route.priority = priority;
    route.block = [handlerBlock copy];
    route.parentRoutesController = self;
    
    if (!route.block) {
        route.block = [^BOOL (NSDictionary *params) {
            return YES;
        } copy];
    }
    
    if (priority == 0 || self.routes.count == 0) {
        [self.routes addObject:route];
    } else {
        NSArray *existingRoutes = self.routes;
        NSUInteger index = 0;
        BOOL addedRoute = NO;
        
        // search through existing routes looking for a lower priority route than this one
        for (_JLRoute *existingRoute in existingRoutes) {
            if (existingRoute.priority < priority) {
                // if found, add the route after it
                [self.routes insertObject:route atIndex:index];
                addedRoute = YES;
                break;
            }
            index++;
        }
        
        // if we weren't able to find a lower priority route, this is the new lowest priority route (or same priority as self.routes.lastObject) and should just be added
        if (!addedRoute)
            [self.routes addObject:route];
    }
}

- (BOOL)_routeURL:(NSURL *)URL withParameters:(NSDictionary *)parameters executeRouteBlock:(BOOL)executeRouteBlock
{
    if (!URL) {
        return NO;
    }
    
    [self _verboseLog:@"Trying to route URL %@", URL];
    
    BOOL didRoute = NO;
    NSDictionary *queryParameters = [URL.query JLRoutes_URLParameterDictionary];
    [self _verboseLog:@"Parsed query parameters: %@", queryParameters];
    
    NSDictionary *fragmentParameters = [URL.fragment JLRoutes_URLParameterDictionary];
    [self _verboseLog:@"Parsed fragment parameters: %@", fragmentParameters];
    
    NSDictionary *fragmentQueryParameters = [URL.JLRoutes_fragmentQuery JLRoutes_URLParameterDictionary];
    [self _verboseLog:@"Parsed fragment query parameters: %@", fragmentParameters];
    
    // break the URL down into path components and filter out any leading/trailing slashes from it
    NSPredicate *filterSlashesPredicate = [NSPredicate predicateWithFormat:@"NOT SELF like '/'"];
    NSArray *pathComponents = [(URL.pathComponents ?: @[]) filteredArrayUsingPredicate:filterSlashesPredicate];
    NSArray *fragmentComponents = [(URL.JLRoutes_fragmentPathComponents ?: @[]) filteredArrayUsingPredicate:filterSlashesPredicate];
    
    if ([URL.host rangeOfString:@"."].location == NSNotFound && ![URL.host isEqualToString:@"localhost"]) {
        // For backward compatibility, handle scheme://path/to/resource as if path was part of the path
        pathComponents = [@[URL.host] arrayByAddingObjectsFromArray:pathComponents];
    }
    
    [self _verboseLog:@"URL path components: %@", pathComponents];
    [self _verboseLog:@"URL fragment components: %@", fragmentComponents];
    
    for (_JLRoute *route in self.routes) {
        NSDictionary *matchParameters = [route parametersForURL:URL pathComponents:pathComponents fragmentComponents:fragmentComponents];
        if (matchParameters) {
            [self _verboseLog:@"Successfully matched %@", route];
            if (!executeRouteBlock) {
                return YES;
            }
            
            // add the URL parameters
            NSMutableDictionary *finalParameters = [NSMutableDictionary dictionary];
            
            // in increasing order of precedence: query, fragment, route, builtin
            [finalParameters addEntriesFromDictionary:queryParameters];
            if (route.matchFragmentComponents) {
                [finalParameters addEntriesFromDictionary:fragmentQueryParameters];
            } else {
                [finalParameters addEntriesFromDictionary:fragmentParameters];
            }
            [finalParameters addEntriesFromDictionary:matchParameters];
            [finalParameters addEntriesFromDictionary:parameters];
            finalParameters[kJLRoutePatternKey] = route.pattern;
            finalParameters[kJLRouteURLKey] = URL;
            __strong __typeof(route.parentRoutesController) strongParentRoutesController = route.parentRoutesController;
            finalParameters[kJLRouteNamespaceKey] = strongParentRoutesController.namespaceKey ?: [NSNull null];
            
            [self _verboseLog:@"Final parameters are %@", finalParameters];
            didRoute = route.block(finalParameters);
            if (didRoute) {
                break;
            }
        }
    }
    
    if (!didRoute) {
        [self _verboseLog:@"Could not find a matching route, returning NO"];
    }
    
    // if we couldn't find a match and this routes controller specifies to fallback and its also not the global routes controller, then...
    if (!didRoute && self.shouldFallbackToGlobalRoutes && ![self _isGlobalRoutesController]) {
        [self _verboseLog:@"Falling back to global routes..."];
        didRoute = [[JLRoutes globalRoutes] _routeURL:URL withParameters:parameters executeRouteBlock:executeRouteBlock];
    }
    
    // if, after everything, we did not route anything and we have an unmatched URL handler, then call it
    if (!didRoute && executeRouteBlock && self.unmatchedURLHandler) {
        self.unmatchedURLHandler(self, URL, parameters);
    }
    
    return didRoute;
}

+ (instancetype)_routesControllerForURL:(NSURL *)URL
{
    if (URL == nil) {
        return nil;
    }
    
    return routeControllersMap[URL.scheme] ?: [JLRoutes globalRoutes];
}

- (BOOL)_isGlobalRoutesController
{
    return [self.namespaceKey isEqualToString:kJLRoutesGlobalNamespaceKey];
}

- (void)_verboseLog:(NSString *)format, ...
{
    if (!verboseLoggingEnabled || format.length == 0) {
        return;
    }
    
    va_list argsList;
    va_start(argsList, format);
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wformat-nonliteral"
    NSString *formattedLogMessage = [[NSString alloc] initWithFormat:format arguments:argsList];
#pragma clang diagnostic pop
    
    va_end(argsList);
    NSLog(@"[JLRoutes]: %@", formattedLogMessage);
}

@end


#pragma mark -

@implementation _JLRoute

- (NSDictionary *)parametersForURL:(NSURL *)URL pathComponents:(NSArray *)pathComponents fragmentComponents:(NSArray *)fragmentComponents
{
    if (!self.patternPathComponents) {
        NSString *fragmentIdentifier = @"#";
        NSRange range = [self.pattern rangeOfString:fragmentIdentifier];

        NSString *pathPattern;
        NSString *fragmentPattern;

        if (range.location != NSNotFound) {
            pathPattern = [self.pattern substringToIndex:range.location];
            fragmentPattern = [self.pattern substringFromIndex:range.location + fragmentIdentifier.length];
            self.matchFragmentComponents = YES;
        } else {
            pathPattern = self.pattern;
        }

        NSPredicate *filterSlashesPredicate = [NSPredicate predicateWithFormat:@"NOT SELF like '/'"];
        self.patternPathComponents = [[pathPattern pathComponents] filteredArrayUsingPredicate:filterSlashesPredicate];
        self.patternFragmentComponents = [[fragmentPattern pathComponents] filteredArrayUsingPredicate:filterSlashesPredicate];
    }

    NSMutableDictionary *routeParameters = [NSMutableDictionary dictionary];

    NSDictionary *pathParameters = [self _parametersForURL:URL patternComponents:self.patternPathComponents components:pathComponents];
    if (!pathParameters) {
        return nil;
    }

    NSDictionary *fragmentParameters;
    if (self.matchFragmentComponents) {
        fragmentParameters = [self _parametersForURL:URL patternComponents:self.patternFragmentComponents components:fragmentComponents];
        if (!fragmentParameters) {
            return nil;
        }

        [routeParameters addEntriesFromDictionary:fragmentParameters];
    }

    [routeParameters addEntriesFromDictionary:pathParameters];


    return routeParameters;
}

- (NSDictionary *)_parametersForURL:(NSURL *)URL patternComponents:(NSArray *)patternComponents components:(NSArray *)components
{
    NSDictionary *routeParameters = nil;

    // do a quick component count check to quickly eliminate incorrect patterns
    BOOL componentCountEqual = patternComponents.count == components.count;
    BOOL patternContainsWildcard = [patternComponents containsObject:@"*"];
    if (componentCountEqual || patternContainsWildcard) {
        // now that we've identified a possible match, move component by component to check if it's a match
        NSUInteger componentIndex = 0;
        NSMutableDictionary *variables = [NSMutableDictionary dictionary];
        BOOL isMatch = YES;

        for (NSString *patternComponent in patternComponents) {
            NSString *URLComponent = nil;
            if (componentIndex < [components count]) {
                URLComponent = components[componentIndex];
            } else if ([patternComponent isEqualToString:@"*"]) { // match /foo by /foo/*
                URLComponent = [components lastObject];
            }

            if ([patternComponent hasPrefix:@":"]) {
                // this component is a variable
                NSString *variableName = [patternComponent substringFromIndex:1];
                NSString *variableValue = URLComponent;
                NSString *urlDecodedVariableValue = [variableValue JLRoutes_URLDecodedString];
                if ([variableName length] > 0 && [urlDecodedVariableValue length] > 0) {
                    variables[variableName] = urlDecodedVariableValue;
                }
                else {
                    NSMutableArray * newComponents = [NSMutableArray arrayWithArray:components];
                    [newComponents addObject:@""];
                    components = newComponents;
                }
            } else if ([patternComponent isEqualToString:@"*"]) {
                // match wildcards
                variables[kJLRouteWildcardComponentsKey] = [components subarrayWithRange:NSMakeRange(componentIndex, components.count-componentIndex)];
                isMatch = YES;
                break;
            } else if (![patternComponent isEqualToString:URLComponent]) {
                // a non-variable component did not match, so this route doesn't match up - on to the next one
                isMatch = NO;
                break;
            }
            componentIndex++;
        }

        if (isMatch) {
            routeParameters = variables;
        }
    }

    return routeParameters;
}

- (NSString *)description
{
    return [NSString stringWithFormat:@"<JLRoute %p> - %@ (priority: %@)", self, self.pattern, @(self.priority)];
}

@end


#pragma mark - Categories

@implementation NSString (JLRoutes)

- (NSString *)JLRoutes_URLDecodedString
{
    NSString *input = shouldDecodePlusSymbols ? [self stringByReplacingOccurrencesOfString:@"+" withString:@" " options:NSLiteralSearch range:NSMakeRange(0, self.length)] : self;
    return [input stringByRemovingPercentEncoding];
}

- (NSDictionary *)JLRoutes_URLParameterDictionary
{
    NSMutableDictionary *parameters = [NSMutableDictionary dictionary];
    
    if (self.length && [self rangeOfString:@"="].location != NSNotFound) {
        NSArray *keyValuePairs = [self componentsSeparatedByString:@"&"];
        for (NSString *keyValuePair in keyValuePairs) {
            NSArray *pair = [keyValuePair componentsSeparatedByString:@"="];
            // don't assume we actually got a real key=value pair. start by assuming we only got @[key] before checking count
            NSString *paramValue = pair.count == 2 ? pair[1] : @"";
            // CFURLCreateStringByReplacingPercentEscapesUsingEncoding may return NULL
            parameters[pair[0]] = [paramValue JLRoutes_URLDecodedString] ?: @"";
        }
    }
    
    return parameters;
}

@end

@implementation NSURL (JLRoutes)

- (NSArray *)JLRoutes_fragmentPathComponents
{
    NSString *fragment = self.fragment;
    if (!fragment) {
        return nil;
    }
    
    NSURL *URL = [[NSURL alloc] initWithString:fragment];
    return URL.pathComponents;
}

- (NSString *)JLRoutes_fragmentQuery
{
    NSString *fragment = self.fragment;
    if (!fragment) {
        return nil;
    }
    
    NSURL *URL = [[NSURL alloc] initWithString:fragment];
    return URL.query;
}

@end


#pragma mark - Global Options

@implementation JLRoutes (GlobalOptions)

+ (void)setVerboseLoggingEnabled:(BOOL)loggingEnabled
{
    verboseLoggingEnabled = loggingEnabled;
}

+ (void)setShouldDecodePlusSymbols:(BOOL)shouldDecode
{
    shouldDecodePlusSymbols = shouldDecode;
}

@end


#pragma mark - Deprecated

@implementation JLRoutes (Deprecated)

+ (void)addRoute:(NSString *)routePattern handler:(BOOL (^)(NSDictionary<NSString *, id> *parameters))handlerBlock
{
    [[self globalRoutes] addRoute:routePattern handler:handlerBlock];
}

+ (void)addRoute:(NSString *)routePattern priority:(NSUInteger)priority handler:(BOOL (^)(NSDictionary<NSString *, id> *parameters))handlerBlock
{
    [[self globalRoutes] addRoute:routePattern priority:priority handler:handlerBlock];
}

+ (void)addRoutes:(NSArray<NSString *> *)routePatterns handler:(BOOL (^)(NSDictionary<NSString *, id> *parameters))handlerBlock
{
    [[self globalRoutes] addRoutes:routePatterns handler:handlerBlock];
}

+ (void)removeRoute:(NSString *)routePattern
{
    [[self globalRoutes] removeRoute:routePattern];
}

+ (void)removeAllRoutes
{
    [[self globalRoutes] removeAllRoutes];
}

+ (BOOL)canRouteURL:(NSURL *)URL withParameters:(NSDictionary *)parameters
{
    return [[self globalRoutes] canRouteURL:URL];
}

- (BOOL)canRouteURL:(NSURL *)URL withParameters:(NSDictionary *)parameters
{
    return [self canRouteURL:URL];
}

@end
